/*
 *  libMirage: Debug context object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mirage.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)

#define SANITY_CHECK(function)                                          \
    if (!self) {                                                        \
        mirage_error(MIRAGE_E_OBJISNULL, error);                        \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_IS_DEBUG_CONTEXT(self)) {                               \
        mirage_error(MIRAGE_E_INVALIDOBJ, error);                       \
        return FALSE;                                                   \
    }                                                                   \
    if (MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self)->dispose_has_run) {      \
        mirage_error(MIRAGE_E_OBJDISPOSED, error);                      \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_DEBUG_CONTEXT_GET_CLASS(self)->function) {              \
        mirage_error(MIRAGE_E_NOFUNCTION, error);                       \
        return FALSE;                                                   \
    }

#define CHECK_ARG(param)                                                \
    if (!param) {                                                       \
        mirage_error(MIRAGE_E_INVALIDARG, error);                       \
        return FALSE;                                                   \
    }


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_DEBUG_CONTEXT, MIRAGE_DebugContextPrivate))

typedef struct _MIRAGE_DebugContextPrivate MIRAGE_DebugContextPrivate;

struct _MIRAGE_DebugContextPrivate {
    gboolean dispose_has_run;
    
    gchar *name; /* Debug context name... e.g. 'Device 1' */
    gchar *domain; /* Debug context domain... e.g. 'libMirage' */
    
    gint debug_mask; /* Debug mask */
};


/******************************************************************************\
 *                 Default implementations of virtual functions               *
\******************************************************************************/
static gboolean __mirage_debug_context_set_domain (MIRAGE_DebugContext *self, gchar *domain, GError **error) {
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    /* Set domain */
    g_free(_priv->domain);
    _priv->domain = g_strdup(domain);
    return TRUE;
}

static gboolean __mirage_debug_context_get_domain (MIRAGE_DebugContext *self, gchar **domain, GError **error) {
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    CHECK_ARG(domain);
    /* Return domain, if it's set */
    if (_priv->domain) {
        *domain = g_strdup(_priv->domain);
    }
    return TRUE;
}

static gboolean __mirage_debug_context_set_name (MIRAGE_DebugContext *self, gchar *name, GError **error) {
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    /* Set name */
    g_free(_priv->name);
    _priv->name = g_strdup(name);
    return TRUE;
}

static gboolean __mirage_debug_context_get_name (MIRAGE_DebugContext *self, gchar **name, GError **error) {
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    CHECK_ARG(name);    
    /* Return name, if it's set */
    if (_priv->name) {
        *name= g_strdup(_priv->name);
    }
    return TRUE;
}

static gboolean __mirage_debug_context_set_debug_mask (MIRAGE_DebugContext *self, gint debug_mask, GError **error) {
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    /* Set debug mask */
    _priv->debug_mask = debug_mask;
    return TRUE;
}

static gboolean __mirage_debug_context_get_debug_mask (MIRAGE_DebugContext *self, gint *debug_mask, GError **error) {
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    CHECK_ARG(debug_mask);
    /* Return debug mask */
    *debug_mask = _priv->debug_mask;
    return TRUE;
}


/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
gboolean mirage_debug_context_set_domain (MIRAGE_DebugContext *self, gchar *domain, GError **error) {
    SANITY_CHECK(set_domain);
    return MIRAGE_DEBUG_CONTEXT_GET_CLASS(self)->set_domain(self, domain, error);
}

gboolean mirage_debug_context_get_domain (MIRAGE_DebugContext *self, gchar **domain, GError **error) {
    SANITY_CHECK(get_domain);
    return MIRAGE_DEBUG_CONTEXT_GET_CLASS(self)->get_domain(self, domain, error);
}

gboolean mirage_debug_context_set_name (MIRAGE_DebugContext *self, gchar *name, GError **error) {
    SANITY_CHECK(set_name);
    return MIRAGE_DEBUG_CONTEXT_GET_CLASS(self)->set_name(self, name, error);
}

gboolean mirage_debug_context_get_name (MIRAGE_DebugContext *self, gchar **name, GError **error) {
    SANITY_CHECK(get_name);
    return MIRAGE_DEBUG_CONTEXT_GET_CLASS(self)->get_name(self, name, error);
}

gboolean mirage_debug_context_set_debug_mask (MIRAGE_DebugContext *self, gint debug_mask, GError **error) {
    SANITY_CHECK(set_debug_mask);
    return MIRAGE_DEBUG_CONTEXT_GET_CLASS(self)->set_debug_mask(self, debug_mask, error);
}

gboolean mirage_debug_context_get_debug_mask (MIRAGE_DebugContext *self, gint *debug_mask, GError **error) {
    SANITY_CHECK(get_debug_mask);
    return MIRAGE_DEBUG_CONTEXT_GET_CLASS(self)->get_debug_mask(self, debug_mask, error);
}


/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static GObjectClass *parent_class = NULL;

static void __mirage_debug_context_dispose (GObject *obj) {
    MIRAGE_DebugContext *self = MIRAGE_DEBUG_CONTEXT(obj);
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
        
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_debug_context_finalize (GObject *obj) {
    MIRAGE_DebugContext *self = MIRAGE_DEBUG_CONTEXT(obj);
    MIRAGE_DebugContextPrivate *_priv = MIRAGE_DEBUG_CONTEXT_GET_PRIVATE(self);
    
    g_free(_priv->domain);
    g_free(_priv->name);
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void __mirage_debug_context_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_DebugContextClass *klass = MIRAGE_DEBUG_CONTEXT_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(MIRAGE_DebugContextPrivate));
    
    /* Initialize GObject members */
    gobject_class->dispose = __mirage_debug_context_dispose;
    gobject_class->finalize = __mirage_debug_context_finalize;
    
    /* Initialize MIRAGE_DebugContext members */
    klass->set_domain = __mirage_debug_context_set_domain;
    klass->get_domain = __mirage_debug_context_get_domain;
    
    klass->set_name = __mirage_debug_context_set_name;
    klass->get_name = __mirage_debug_context_get_name;
    
    klass->set_debug_mask = __mirage_debug_context_set_debug_mask;
    klass->get_debug_mask = __mirage_debug_context_get_debug_mask;
    
    return;
}

GType mirage_debug_context_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_DebugContextClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_debug_context_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_DebugContext),
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(G_TYPE_OBJECT, "MIRAGE_DebugContext", &info, 0);
    }
    
    return type;
}
