/*
 *  libMirage: Fragment object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mirage.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)

#define SANITY_CHECK(function)                                          \
    if (!self) {                                                        \
        mirage_error(MIRAGE_E_OBJISNULL, error);                        \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_IS_FRAGMENT(self)) {                                    \
        mirage_error(MIRAGE_E_INVALIDOBJ, error);                       \
        return FALSE;                                                   \
    }                                                                   \
    if (MIRAGE_FRAGMENT_GET_PRIVATE(self)->dispose_has_run) {           \
        mirage_error(MIRAGE_E_OBJDISPOSED, error);                      \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_FRAGMENT_GET_CLASS(self)->function) {                   \
        mirage_error(MIRAGE_E_NOFUNCTION, error);                       \
        return FALSE;                                                   \
    }

#define CHECK_ARG(param)                                                \
    if (!param) {                                                       \
        mirage_error(MIRAGE_E_INVALIDARG, error);                       \
        return FALSE;                                                   \
    }


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_FRAGMENT_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_FRAGMENT, MIRAGE_FragmentPrivate))

typedef struct {
    gboolean dispose_has_run;
        
    gint address;    /* Address (relative to track start) */
    gint length;     /* Length, in sectors */
} MIRAGE_FragmentPrivate;


/******************************************************************************\
 *                              Private functions                             *
\******************************************************************************/
static gboolean __mirage_fragment_commit_topdown_change (MIRAGE_Fragment *self, GError **error) {
    /* Nothing to do here */
    return TRUE;
}

static gboolean __mirage_fragment_commit_bottomup_change (MIRAGE_Fragment *self, GError **error) {
    /* Signal fragment change */
    g_signal_emit_by_name(MIRAGE_OBJECT(self), "object-modified", NULL);
    
    return TRUE;
}

/******************************************************************************\
 *                 Default implementations of virtual functions               *
\******************************************************************************/
static gboolean __mirage_fragment_set_address (MIRAGE_Fragment *self, gint address, GError **error) {
    MIRAGE_FragmentPrivate *_priv = MIRAGE_FRAGMENT_GET_PRIVATE(self);
    /* Set address */
    _priv->address = address;
    /* Top-down change */
    __mirage_fragment_commit_topdown_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_fragment_get_address (MIRAGE_Fragment *self, gint *address, GError **error) {
    MIRAGE_FragmentPrivate *_priv = MIRAGE_FRAGMENT_GET_PRIVATE(self);
    CHECK_ARG(address);
    /* Return address */
    *address = _priv->address;
    return TRUE;
}

    
static gboolean __mirage_fragment_set_length (MIRAGE_Fragment *self, gint length, GError **error) {
    MIRAGE_FragmentPrivate *_priv = MIRAGE_FRAGMENT_GET_PRIVATE(self);
    /* Set length */
    _priv->length = length;
    /* Bottom-up change */
    __mirage_fragment_commit_bottomup_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_fragment_get_length (MIRAGE_Fragment *self, gint *length, GError **error) {
    MIRAGE_FragmentPrivate *_priv = MIRAGE_FRAGMENT_GET_PRIVATE(self);
    CHECK_ARG(length);
    /* Return length */
    *length = _priv->length;
    return TRUE;
}


/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
gboolean mirage_fragment_get_fragment_type (MIRAGE_Fragment *self, gchar **fragment_type, GError **error) {
    SANITY_CHECK(get_fragment_type);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->get_fragment_type(self, fragment_type, error);
}

gboolean mirage_fragment_set_address (MIRAGE_Fragment *self, gint address, GError **error) {
    SANITY_CHECK(set_address);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->set_address(self, address, error);
}

gboolean mirage_fragment_get_address (MIRAGE_Fragment *self, gint *address, GError **error) {
    SANITY_CHECK(get_address);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->get_address(self, address, error);
}

    
gboolean mirage_fragment_set_length (MIRAGE_Fragment *self, gint length, GError **error) {
    SANITY_CHECK(set_length);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->set_length(self, length, error);
}

gboolean mirage_fragment_get_length (MIRAGE_Fragment *self, gint *length, GError **error) {
    SANITY_CHECK(get_length);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->get_length(self, length, error);
}


gboolean mirage_fragment_use_the_rest_of_file (MIRAGE_Fragment *self, GError **error) {
    SANITY_CHECK(use_the_rest_of_file);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->use_the_rest_of_file(self, error);
}

    
gboolean mirage_fragment_read_main_data (MIRAGE_Fragment *self, gint address, guint8 *buf, gint *length, GError **error) {
    SANITY_CHECK(read_main_data);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->read_main_data(self, address, buf, length, error);
}

gboolean mirage_fragment_read_subchannel_data (MIRAGE_Fragment *self, gint address, guint8 *buf, gint *length, GError **error) {
    SANITY_CHECK(read_subchannel_data);
    return MIRAGE_FRAGMENT_GET_CLASS(self)->read_subchannel_data(self, address, buf, length, error);
}


/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static MIRAGE_ObjectClass *parent_class = NULL;

static void __mirage_fragment_dispose (GObject *obj) {
    MIRAGE_Fragment *self = MIRAGE_FRAGMENT(obj);
    MIRAGE_FragmentPrivate *_priv = MIRAGE_FRAGMENT_GET_PRIVATE(self);
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
        
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_fragment_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_FragmentClass *klass = MIRAGE_FRAGMENT_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(MIRAGE_FragmentPrivate));
    
    /* Initialize GObject members */
    gobject_class->dispose = __mirage_fragment_dispose;
    
    /* Initialize MIRAGE_Fragment members */    
    klass->set_address = __mirage_fragment_set_address;
    klass->get_address = __mirage_fragment_get_address;
    
    klass->set_length = __mirage_fragment_set_length;
    klass->get_length = __mirage_fragment_get_length;
    
    /* Following functions are provided by implementations */
    klass->get_fragment_type = NULL;
    klass->use_the_rest_of_file = NULL;
    klass->read_main_data = NULL;
    klass->read_subchannel_data = NULL;
    
    return;
}

GType mirage_fragment_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_FragmentClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_fragment_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Fragment),
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_OBJECT, "MIRAGE_Fragment", &info, 0);
    }
    
    return type;
}

/******************************************************************************\
 *                                NULL interface                              *
\******************************************************************************/
GType mirage_finterface_null_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_FInterface_NULLClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            NULL,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            0,
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(G_TYPE_INTERFACE, "MIRAGE_FInterface_NULL", &info, 0);
    }
    
    return type;
}


/******************************************************************************\
 *                               BINARY interface                             *
\******************************************************************************/
gboolean mirage_finterface_binary_track_file_set_handle (MIRAGE_FInterface_BINARY *self, FILE *file, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_set_handle(self, file, error);
}

gboolean mirage_finterface_binary_track_file_get_handle (MIRAGE_FInterface_BINARY *self, FILE **file, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_get_handle(self, file, error);
}

gboolean mirage_finterface_binary_track_file_set_offset (MIRAGE_FInterface_BINARY *self, guint64 offset, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_set_offset(self, offset, error);
}

gboolean mirage_finterface_binary_track_file_get_offset (MIRAGE_FInterface_BINARY *self, guint64 *offset, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_get_offset(self, offset, error);
}


gboolean mirage_finterface_binary_track_file_set_sectsize (MIRAGE_FInterface_BINARY *self, gint sectsize, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_set_sectsize(self, sectsize, error);    
}

gboolean mirage_finterface_binary_track_file_get_sectsize (MIRAGE_FInterface_BINARY *self, gint *sectsize, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_get_sectsize(self, sectsize, error);    
}


gboolean mirage_finterface_binary_track_file_set_format (MIRAGE_FInterface_BINARY *self, gint format, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_set_format(self, format, error);    
}

gboolean mirage_finterface_binary_track_file_get_format (MIRAGE_FInterface_BINARY *self, gint *format, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_get_format(self, format, error);        
}


gboolean mirage_finterface_binary_track_file_get_position (MIRAGE_FInterface_BINARY *self, gint address, guint64 *position, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->track_file_get_position(self, address, position, error);            
}


gboolean mirage_finterface_binary_subchannel_file_set_handle (MIRAGE_FInterface_BINARY *self, FILE *file, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_set_handle(self, file, error);
}

gboolean mirage_finterface_binary_subchannel_file_get_handle (MIRAGE_FInterface_BINARY *self, FILE **file, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_get_handle(self, file, error);
}


gboolean mirage_finterface_binary_subchannel_file_set_offset (MIRAGE_FInterface_BINARY *self, guint64 offset, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_set_offset(self, offset, error);
}

gboolean mirage_finterface_binary_subchannel_file_get_offset (MIRAGE_FInterface_BINARY *self, guint64 *offset, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_get_offset(self, offset, error);
}

gboolean mirage_finterface_binary_subchannel_file_set_sectsize (MIRAGE_FInterface_BINARY *self, gint sectsize, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_set_sectsize(self, sectsize, error);
}

gboolean mirage_finterface_binary_subchannel_file_get_sectsize (MIRAGE_FInterface_BINARY *self, gint *sectsize, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_get_sectsize(self, sectsize, error);
}


gboolean mirage_finterface_binary_subchannel_file_set_format (MIRAGE_FInterface_BINARY *self, gint format, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_set_format(self, format, error);
}

gboolean mirage_finterface_binary_subchannel_file_get_format (MIRAGE_FInterface_BINARY *self, gint *format, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_get_format(self, format, error);
}


gboolean mirage_finterface_binary_subchannel_file_get_position (MIRAGE_FInterface_BINARY *self, gint address, guint64 *position, GError **error) {
    return MIRAGE_FINTERFACE_BINARY_GET_CLASS(self)->subchannel_file_get_position(self, address, position, error);
}


GType mirage_finterface_binary_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_FInterface_BINARYClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            NULL,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            0,
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(G_TYPE_INTERFACE, "MIRAGE_FInterface_BINARYClass", &info, 0);
    }
    
    return type;
}


/******************************************************************************\
 *                                AUDIO interface                             *
\******************************************************************************/
gboolean mirage_finterface_audio_set_file (MIRAGE_FInterface_AUDIO *self, gchar *filename, GError **error) {
    return MIRAGE_FINTERFACE_AUDIO_GET_CLASS(self)->set_file(self, filename, error);
}

gboolean mirage_finterface_audio_set_offset (MIRAGE_FInterface_AUDIO *self, gint offset, GError **error) {
    return MIRAGE_FINTERFACE_AUDIO_GET_CLASS(self)->set_offset(self, offset, error);
}


GType mirage_finterface_audio_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_FInterface_AUDIOClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            NULL,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            0,
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(G_TYPE_INTERFACE, "MIRAGE_FInterface_AUDIOClass", &info, 0);
    }
    
    return type;
}
