/*
 *  libMirage: Mirage object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mirage.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
#define CHECK_ARG(param)                                            \
    if (!param) {                                                   \
        mirage_error(MIRAGE_E_INVALIDARG, error);                   \
        return FALSE;                                               \
    }

/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_MIRAGE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_MIRAGE, MIRAGE_MiragePrivate))

typedef struct {
    gboolean dispose_has_run;
    
    gchar *version;
    
    GList *plugins_list_parsers; /* Parser plugins */
    GList *plugins_list_fragments; /* Fragment plugins */
} MIRAGE_MiragePrivate;


/******************************************************************************\
 *                              Private functions                             *
\******************************************************************************/
static void __mirage_mirage_load_plugins (MIRAGE_Mirage *self) {
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);

    GDir *plugins_dir = g_dir_open(MIRAGE_PLUGIN_DIR, 0, NULL);
    
    if (plugins_dir != NULL) {
        const gchar *plugin_file = NULL;
        while ((plugin_file = g_dir_read_name(plugins_dir)) != NULL) {
            if (g_str_has_suffix(plugin_file, ".so")) {
                MIRAGE_PluginEntry *new_entry = NULL;
                new_entry = g_new0(MIRAGE_PluginEntry, 1);

                /* Build full path */
                gchar *fullpath = g_build_filename(MIRAGE_PLUGIN_DIR, plugin_file, NULL);
                /* Load */
                new_entry->handle = g_module_open(fullpath, G_MODULE_BIND_MASK);
                g_module_make_resident(new_entry->handle); /* Make sure module isn't unloaded until lib is closed */
                if (new_entry->handle) {
                    gboolean succeeded = FALSE;
                    succeeded = g_module_symbol(new_entry->handle, "mirage_plugin_info", (void**)&new_entry->info);
                    if (succeeded) {
                        switch (new_entry->info->type) {
                            case MIRAGE_PLUGIN_FRAGMENT: {
                                _priv->plugins_list_fragments = g_list_append(_priv->plugins_list_fragments, new_entry);
                                break;
                            }
                            case MIRAGE_PLUGIN_PARSER: {
                                _priv->plugins_list_parsers = g_list_append(_priv->plugins_list_parsers, new_entry);
                                break;
                            }
                            default: {
                                MIRAGE_DEBUG(self, MIRAGE_DEBUG_WARNING, "%s: unknown plugin type: %s\n", __func__, fullpath);
                                g_free(new_entry);
                                break;
                            }
                        }
                    } else {
                        MIRAGE_DEBUG(self, MIRAGE_DEBUG_WARNING, "%s: failed to get plugin_info from module '%s'!\n", __func__, plugin_file);
                    }
                } else {
                    MIRAGE_DEBUG(self, MIRAGE_DEBUG_WARNING, "%s: failed to open handle for plugin '%s'!\n", __func__, plugin_file);
                }                    
                /* Cleanup */
                g_free(fullpath);
            }
        }
        
        g_dir_close(plugins_dir);
    }
}


/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
gboolean mirage_mirage_get_version (MIRAGE_Mirage *self, gchar **version, GError **error) {
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);
    
    CHECK_ARG(version);
    *version = g_strdup(_priv->version);
    
    return TRUE;
}

gboolean mirage_mirage_create_disc (MIRAGE_Mirage *self, gchar **filenames, GObject **ret_disc, GObject *debug_context, GError **error) {
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);

    GList *entry = NULL;
    gboolean ret = FALSE;
    
    gint i;
    
    CHECK_ARG(filenames);
    CHECK_ARG(ret_disc);
        
    /* Check if filename(s) is/are valid */
    for (i = 0; i < g_strv_length(filenames); i++) {
        if (!g_file_test(filenames[i], G_FILE_TEST_IS_REGULAR)) {
            mirage_error(MIRAGE_E_IMAGEFILE, error);
            return FALSE;
        }
    }
    
    /* Find appropriate plugin... for now it's required that one plugin handles
       all given files */
    G_LIST_FOR_EACH(entry, _priv->plugins_list_parsers) {
        MIRAGE_PluginEntry *plugin_entry = (MIRAGE_PluginEntry *)entry->data;
        MIRAGE_PluginInfo_Parser *parser_info = (MIRAGE_PluginInfo_Parser *)plugin_entry->info;

        /* Check whether given plugin can handle given filename(s) */
        gboolean can_handle_one_image = FALSE;
        for (i = 0; i < g_strv_length(filenames); i++) {
            ret = parser_info->can_handle_image(filenames[i]);
            /* We break the loop in case previous filename could be handled,
               but this one cannot be */
            if (can_handle_one_image && !ret) {
                break;
            }
            can_handle_one_image = TRUE;
        }
        
        /* If the plugin can handle all of 'em, go ahead and make disc */
        if (ret) {
            /* Plug-in can handle image, create disc */
            GObject *disc = parser_info->create_disc();
            /* Set Mirage (= self) to disc */
            mirage_object_set_mirage(MIRAGE_OBJECT(disc), G_OBJECT(self), NULL);
            /* If provided, attach the debug context to disc */
            if (debug_context) {
                mirage_object_set_debug_context(MIRAGE_OBJECT(disc), debug_context, NULL);
            }
            /* Try loading image */
            ret = mirage_disc_load_image(MIRAGE_DISC(disc), filenames, error);
            if (!ret) {
                /* Loading failed; delete disc and return false */
                g_object_unref(disc);
                return FALSE;
            }
            
            *ret_disc = disc;
            return TRUE;
        }
    }
    
    *ret_disc = NULL;

    mirage_error(MIRAGE_E_NOPARSERFOUND, error);
    return FALSE;
}


gboolean mirage_mirage_create_fragment (MIRAGE_Mirage *self, gint fragment_interface, gchar *filename, GObject **ret_fragment, GError **error) {
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);

    GList *entry = NULL;
        
    CHECK_ARG(filename);
    CHECK_ARG(ret_fragment);
        
    /* Find appropriate plugin... */
    G_LIST_FOR_EACH(entry, _priv->plugins_list_fragments) {
        MIRAGE_PluginEntry *plugin_entry = (MIRAGE_PluginEntry *)entry->data;
        MIRAGE_PluginInfo_Fragment *parser_info = (MIRAGE_PluginInfo_Fragment *)plugin_entry->info;

        /* Check whether given plugin can handle given file */
        if (parser_info->can_handle_data(filename, fragment_interface)) {
             /* Plug-in can handle image, create disc */
            GObject *fragment = parser_info->create_fragment();
            /* Set Mirage (= self) to disc */
            mirage_object_set_mirage(MIRAGE_OBJECT(fragment), G_OBJECT(self), NULL);
                        
            *ret_fragment = fragment;
            return TRUE;
        }
    }
    
    *ret_fragment = NULL;

    mirage_error(MIRAGE_E_NOFRAGMENTFOUND, error);
    return FALSE;
}


gboolean mirage_mirage_for_each_plugin (MIRAGE_Mirage *self, gint plugin_type, GFunc func, gpointer user_data, GError **error) {
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);
    GList *entry = NULL;
    
    CHECK_ARG(func);
    
    /* Parser plugins */
    if (plugin_type & MIRAGE_PLUGIN_PARSER) {
        G_LIST_FOR_EACH(entry, _priv->plugins_list_parsers) {
            MIRAGE_PluginEntry *entry_data = entry->data;
            (*func) (entry_data->info, user_data);
        }
    }
    
    /* Fragment plugins */
    if (plugin_type & MIRAGE_PLUGIN_FRAGMENT) {
        G_LIST_FOR_EACH(entry, _priv->plugins_list_fragments) {
            MIRAGE_PluginEntry *entry_data = entry->data;
            (*func) (entry_data->info, user_data);
        }
    }
    
    return TRUE;
}

gboolean mirage_mirage_get_supported_debug_masks (MIRAGE_Mirage *self, GPtrArray **masks, GError **error) {
    gint i = 0;
    
    static const struct {
        gint mask_value;
        gchar *mask_name;
    } dbg_masks[] = {
        { MIRAGE_DEBUG_CHAIN, "MIRAGE_DEBUG_CHAIN" },
        { MIRAGE_DEBUG_PARSER, "MIRAGE_DEBUG_PARSER" },
        { MIRAGE_DEBUG_DISC, "MIRAGE_DEBUG_DISC" },
        { MIRAGE_DEBUG_SESSION, "MIRAGE_DEBUG_SESSION" },
        { MIRAGE_DEBUG_TRACK, "MIRAGE_DEBUG_TRACK" },
        { MIRAGE_DEBUG_SECTOR, "MIRAGE_DEBUG_SECTOR" },
        { MIRAGE_DEBUG_FRAGMENT, "MIRAGE_DEBUG_FRAGMENT" },
    };
    
    CHECK_ARG(masks);
    
    *masks = g_ptr_array_new();
    
    
    for (i = 0; i < G_N_ELEMENTS(dbg_masks); i++) {
        /* Create value array */
        GValueArray *mask_entry = g_value_array_new(2);
        /* Mask name */
        g_value_array_append(mask_entry, NULL);
        g_value_init(g_value_array_get_nth(mask_entry, 0), G_TYPE_STRING);
        g_value_set_string(g_value_array_get_nth(mask_entry, 0), dbg_masks[i].mask_name);
        /* Mask value */
        g_value_array_append(mask_entry, NULL);
        g_value_init(g_value_array_get_nth(mask_entry, 1), G_TYPE_INT);
        g_value_set_int(g_value_array_get_nth(mask_entry, 1), dbg_masks[i].mask_value);
        /* Add mask's value array to masks' pointer array */
        g_ptr_array_add(*masks, mask_entry);
    }
    
    return TRUE;
}

/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static MIRAGE_ObjectClass *parent_class = NULL;

static void __mirage_mirage_instance_init (GTypeInstance *instance, gpointer g_class) {
    MIRAGE_Mirage *self = MIRAGE_MIRAGE(instance);
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);
    
    /* Set version string */
    _priv->version = g_strdup(PACKAGE_VERSION);
    /* Load Plug-Ins */
    __mirage_mirage_load_plugins(self);

    return;
}

static void __mirage_mirage_dispose (GObject *obj) {
    MIRAGE_Mirage *self = MIRAGE_MIRAGE(obj);
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
    
    /* Unref all members we own references to */
    /* Clear plug-in list */
    GList *entry = NULL;
    G_LIST_FOR_EACH(entry, _priv->plugins_list_fragments) {
        if (entry->data) {
            MIRAGE_PluginEntry *p_entry = (MIRAGE_PluginEntry *)entry->data;
            g_module_close(p_entry->handle);
            g_free(entry->data);
        }
    }
    G_LIST_FOR_EACH(entry, _priv->plugins_list_parsers) {
        if (entry->data) {
            MIRAGE_PluginEntry *p_entry = (MIRAGE_PluginEntry *)entry->data;
            g_module_close(p_entry->handle);
            g_free(entry->data);
        }
    }
    
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_mirage_finalize (GObject *obj) {
    MIRAGE_Mirage *self = MIRAGE_MIRAGE(obj);
    MIRAGE_MiragePrivate *_priv = MIRAGE_MIRAGE_GET_PRIVATE(self);
    
    /* Free version string */
    g_free(_priv->version);
    /* Free plugins list */
    g_list_free(_priv->plugins_list_fragments);
    g_list_free(_priv->plugins_list_parsers);
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void __mirage_mirage_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_MirageClass *klass = MIRAGE_MIRAGE_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(MIRAGE_MiragePrivate));
    
    /* Initialize GObject methods */
    gobject_class->dispose = __mirage_mirage_dispose;
    gobject_class->finalize = __mirage_mirage_finalize;
    
    return;
}

GType mirage_mirage_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_MirageClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_mirage_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Mirage),
            0,      /* n_preallocs */
            __mirage_mirage_instance_init    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_OBJECT, "MIRAGE_Mirage", &info, 0);
    }
    
    return type;
}
