/*
 *  libMirage: Session object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mirage.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)

#define SANITY_CHECK(function)                                      \
    if (!self) {                                                    \
        mirage_error(MIRAGE_E_OBJISNULL, error);                    \
        return FALSE;                                               \
    }                                                               \
    if (!MIRAGE_IS_SESSION(self)) {                                 \
        mirage_error(MIRAGE_E_INVALIDOBJ, error);                   \
        return FALSE;                                               \
    }                                                               \
    if (MIRAGE_SESSION_GET_PRIVATE(self)->dispose_has_run) {        \
        mirage_error(MIRAGE_E_OBJDISPOSED, error);                  \
        return FALSE;                                               \
    }                                                               \
    if (!MIRAGE_SESSION_GET_CLASS(self)->function) {                \
        mirage_error(MIRAGE_E_NOFUNCTION, error);                   \
        return FALSE;                                               \
    }

#define CHECK_ARG(param)                                            \
    if (!param) {                                                   \
        mirage_error(MIRAGE_E_INVALIDARG, error);                   \
        return FALSE;                                               \
    }


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_SESSION_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_SESSION, MIRAGE_SessionPrivate))

typedef struct {
    gboolean dispose_has_run;
    
    /* Layout settings */
    gint session_number; /* Session number */
    gint start_sector;   /* Start sector */
    gint first_track;    /* Number of the first track in session */
    gint length;         /* Length of session (sum of tracks' length) */
    
    /* Session type */
    gint session_type;

    /* Tracks list */
    GList *tracks_list;
    
    /* CD-Text list */
    GList *languages_list;
} MIRAGE_SessionPrivate;


/******************************************************************************\
 *                              Private functions                             *
\******************************************************************************/
/* Forward declarations */
static gboolean __mirage_session_commit_topdown_change (MIRAGE_Session *self, GError **error);
static gboolean __mirage_session_commit_bottomup_change (MIRAGE_Session *self, GError **error);

static void __track_modified_handler (GObject *track, MIRAGE_Session *self);

static gboolean __remove_track_from_session (MIRAGE_Session *self, GObject *track, GError **error);
static gboolean __remove_language_from_session (MIRAGE_Session *self, GObject *language, GError **error);

static gint __sort_languages_by_code (GObject *language1, GObject *language2);
static gint __sort_tracks_by_number (GObject *track1, GObject *track2);


static gboolean __mirage_session_commit_topdown_change (MIRAGE_Session *self, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GList *entry = NULL;
        
    /* Rearrange tracks: set numbers, set start sectors */
    gint cur_track_address = _priv->start_sector;
    gint cur_track_number  = _priv->first_track;
    
    G_LIST_FOR_EACH(entry, _priv->tracks_list) {
        GObject *track = entry->data;
        
        /* Set track's number */
        gint old_number = 0;
        mirage_track_layout_get_track_number(MIRAGE_TRACK(track), &old_number, NULL);
        if ((old_number != MIRAGE_TRACK_LEADIN) && (old_number != MIRAGE_TRACK_LEADOUT)) {
            mirage_track_layout_set_track_number(MIRAGE_TRACK(track), cur_track_number, NULL);
            cur_track_number++;
        }
        
        /* Set track's start address */
        mirage_track_layout_set_start_sector(MIRAGE_TRACK(track), cur_track_address, NULL);
        gint track_length = 0;
        mirage_track_layout_get_length(MIRAGE_TRACK(track), &track_length, NULL);
        cur_track_address += track_length;
    }
        
    return TRUE;
}

static gboolean __mirage_session_commit_bottomup_change (MIRAGE_Session *self, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GList *entry = NULL;
        
    /* Calculate session length */
    _priv->length = 0; /* Reset; it'll be recalculated */

    G_LIST_FOR_EACH(entry, _priv->tracks_list) {
        GObject *track = entry->data;
        gint track_length = 0;
        mirage_track_layout_get_length(MIRAGE_TRACK(track), &track_length, NULL);
        _priv->length += track_length;
    }
    
    /* Signal session change */
    g_signal_emit_by_name(MIRAGE_OBJECT(self), "object-modified", NULL);
    /* If we don't have parent, we should complete the arc by committing top-down change */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(self), NULL, NULL)) {
        __mirage_session_commit_topdown_change(self, NULL);
    }
    
    return TRUE;
}

static void __track_modified_handler (GObject *track, MIRAGE_Session *self) {    
    /* Bottom-up change */
    __mirage_session_commit_bottomup_change(self, NULL);
    return;
}

static gboolean __remove_track_from_session (MIRAGE_Session *self, GObject *track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
                
    /* Disconnect signal handler (find it by handler function and user data) */
    g_signal_handlers_disconnect_by_func(MIRAGE_OBJECT(track), __track_modified_handler, self);
    
    /* Remove track from list and unref it */
    _priv->tracks_list = g_list_remove(_priv->tracks_list, track);
    g_object_unref(track);
    
    /* Bottom-up change */
    __mirage_session_commit_bottomup_change(self, NULL);
    
    return TRUE;
}

static gboolean __remove_language_from_session (MIRAGE_Session *self, GObject *language, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
            
    /* Remove it from list and unref it */
    _priv->languages_list = g_list_remove(_priv->languages_list, language);
    g_object_unref(language);
    
    return TRUE;
}

static gint __sort_languages_by_code (GObject *language1, GObject *language2) {   
    gint code1 = 0;
    gint code2 = 0;
    
    mirage_language_get_langcode(MIRAGE_LANGUAGE(language1), &code1, NULL);
    mirage_language_get_langcode(MIRAGE_LANGUAGE(language2), &code2, NULL);
    
    if (code1 < code2) {
        return -1;
    } else if (code1 > code2) {
        return 1;
    } else {
        return 0;
    }
}

static gint __sort_tracks_by_number (GObject *track1, GObject *track2) {          
    gint number1 = 0;
    gint number2 = 0;
        
    mirage_track_layout_get_track_number(MIRAGE_TRACK(track1), &number1, NULL);
    mirage_track_layout_get_track_number(MIRAGE_TRACK(track2), &number2, NULL);
    
    if (number1 == MIRAGE_TRACK_LEADIN) {
        /* Track 1 is lead-in; always before the rest */
        return -1;
    } else if (number2 == MIRAGE_TRACK_LEADIN) {
        /* Track 2 is lead-in; place Track 1 behind it */
        return 1;
    } else if (number1 == MIRAGE_TRACK_LEADOUT) {
        /* Track 1 is lead-out; always after the rest */
        return 1;
    } else if (number2 == MIRAGE_TRACK_LEADOUT) {
        /* Track 2 is lead-out; place Track 1 before it */
        return -1;
    } else {
        if (number1 < number2) {
            return -1;
        } else if (number1 > number2) {
            return 1;
        } else {
            return 0;
        }
    }
}



/******************************************************************************\
 *                 Default implementations of virtual functions               *
\******************************************************************************/
static gboolean __mirage_session_set_session_type (MIRAGE_Session *self, gint type, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    /* Set session type */
    _priv->session_type =type;
    return TRUE;
}

static gboolean __mirage_session_get_session_type (MIRAGE_Session *self, gint *type, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    CHECK_ARG(type);
    /* Return session number */
    *type = _priv->session_type;
    return TRUE;
}


static gboolean __mirage_session_layout_set_session_number (MIRAGE_Session *self, gint number, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    /* Set session number */
    _priv->session_number = number;
    return TRUE;
}

static gboolean __mirage_session_layout_get_session_number (MIRAGE_Session *self, gint *number, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    CHECK_ARG(number);
    /* Return session number */
    *number = _priv->session_number;
    return TRUE;
}

static gboolean __mirage_session_layout_set_first_track (MIRAGE_Session *self, gint first_track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    /* Set first track */
    _priv->first_track = first_track;
    /* Top-down change */
    __mirage_session_commit_topdown_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_session_layout_get_first_track (MIRAGE_Session *self, gint *first_track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    CHECK_ARG(first_track);
    /* Return first track */
    *first_track = _priv->first_track;
    return TRUE;
}

static gboolean __mirage_session_layout_set_start_sector (MIRAGE_Session *self, gint start_sector, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    /* Set start sector */
    _priv->start_sector = start_sector;
    /* Top-down change */
    __mirage_session_commit_topdown_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_session_layout_get_start_sector (MIRAGE_Session *self, gint *start_sector, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    CHECK_ARG(start_sector);
    /* Return start sector */
    *start_sector = _priv->start_sector;
    return TRUE;
}

static gboolean __mirage_session_layout_get_length (MIRAGE_Session *self, gint *length, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    CHECK_ARG(length);
    /* Return length */
    *length = _priv->length;
    return TRUE;
}


static gboolean __mirage_session_get_number_of_tracks (MIRAGE_Session *self, gint *number_of_tracks, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    CHECK_ARG(number_of_tracks);
    /* Return number of tracks */
    *number_of_tracks = g_list_length(_priv->tracks_list) - 2; /* Length of list, without lead-in and lead-out */
    return TRUE;
}

static gboolean __mirage_session_add_track_by_index (MIRAGE_Session *self, gint index, GObject **track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *new_track = NULL;
    gint num_tracks = 0;
    
    /* First track, last track... allow negative indexes to go from behind */
    mirage_session_get_number_of_tracks(self, &num_tracks, NULL);
    if (index < -num_tracks) {
        /* If negative index is too big, return the first track */
        index = 0;
    }
    if (index > num_tracks) {
        /* If positive index is too big, return the last track */
        index = num_tracks;
    }
    if (index < 0) {
        index += num_tracks + 1;
    }
    
    /* If there's track provided, use it; else create new track */
    if (track && *track) {
        new_track = *track;
        /* If track is not MIRAGE_Track... */
        if (!MIRAGE_IS_TRACK(new_track)) {
            mirage_error(MIRAGE_E_INVALIDOBJTYPE, error);
            return FALSE;
        }
        g_object_ref(new_track);
    } else {
        new_track = g_object_new(MIRAGE_TYPE_TRACK, NULL);
    }

    /* We don't set track number here, because layout recalculation will do it for us */
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_track), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_track, NULL);
    
    /* Insert track into track list... take into account that lead-in has index 0,
       thus all indexes should be increased by 1 */
    _priv->tracks_list = g_list_insert(_priv->tracks_list, new_track, index + 1);
    
    /* Connect track modified signal */
    g_signal_connect(MIRAGE_OBJECT(new_track), "object-modified", (GCallback)__track_modified_handler, self);
    
    /* Bottom-up change */
    __mirage_session_commit_bottomup_change(self, NULL);
    
    /* Return track to user if she wants it */
    if (track && (*track == NULL)) {
        g_object_ref(new_track);
        *track = new_track;
    }
    
    return TRUE;
}

static gboolean __mirage_session_add_track_by_number (MIRAGE_Session *self, gint number, GObject **track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *new_track = NULL;
    
    /* Check if track with that number already exists */
    if (mirage_session_get_track_by_number(self, number, NULL, NULL)) {
        mirage_error(MIRAGE_E_TRACKEXISTS, error);
        return FALSE;
    }
    
    /* If there's track provided, use it; else create new one */
    if (track && *track) {
        new_track = *track;
        /* If track is not MIRAGE_Track... */
        if (!MIRAGE_IS_TRACK(new_track)) {
            mirage_error(MIRAGE_E_INVALIDOBJTYPE, error);
            return FALSE;
        }
        g_object_ref(new_track);
    } else {
        new_track = g_object_new(MIRAGE_TYPE_TRACK, NULL);
    }
    
    /* Set track number */
    mirage_track_layout_set_track_number(MIRAGE_TRACK(new_track), number, NULL);
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_track), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_track, NULL);
    
    /* Insert track into track list */
    _priv->tracks_list = g_list_insert_sorted(_priv->tracks_list, new_track, (GCompareFunc)__sort_tracks_by_number);
    
    /* Connect track modified signal */
    g_signal_connect(MIRAGE_OBJECT(new_track), "object-modified", (GCallback)__track_modified_handler, self);
    
    /* Bottom-up change */
    __mirage_session_commit_bottomup_change(self, NULL);
    
    /* Return track to user if she wants it */
    if (track && (*track == NULL)) {
        g_object_ref(new_track);
        *track = new_track;
    }
    
    return TRUE;
}

static gboolean __mirage_session_remove_track_by_index (MIRAGE_Session *self, gint index, GError **error) {
    /*MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);*/
    GObject *track = NULL;
    gboolean succeeded = FALSE;
    
    /* Find track by index */
    if (!mirage_session_get_track_by_index(self, index, &track, error)) {
        return FALSE;
    }
    
    /* Remove track from list */
    succeeded = __remove_track_from_session(self, track, error);
    g_object_unref(track); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_session_remove_track_by_number (MIRAGE_Session *self, gint number, GError **error) {
    /*MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);*/
    GObject *track = NULL;
    gboolean succeeded = FALSE;
    
    /* You can't delete lead-in/lead-out */
    if (number == MIRAGE_TRACK_LEADIN || number == MIRAGE_TRACK_LEADOUT) {
        mirage_error(MIRAGE_E_INVALIDARG, error);
        return FALSE;
    }
        
    /* Find track in layout */
    if (!mirage_session_get_track_by_number(self, number, &track, error)) {
        return FALSE;
    }
    
    /* Remove track from list */
    succeeded = __remove_track_from_session(self, track, error);
    g_object_unref(track); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_session_remove_track_by_object (MIRAGE_Session *self, GObject *track, GError **error) {
    CHECK_ARG(track);
    return __remove_track_from_session(self, track, error);
}

static gboolean __mirage_session_get_track_by_index (MIRAGE_Session *self, gint index, GObject **track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *ret_track = NULL;
    gint num_tracks = 0;
    
    /* First track, last track... allow negative indexes to go from behind */
    mirage_session_get_number_of_tracks(self, &num_tracks, NULL);
    if (index < -num_tracks || index >= num_tracks) {
        mirage_error(MIRAGE_E_INDEXOUTOFRANGE, error);
        return FALSE;
    } else if (index < 0) {
        index += num_tracks; 
    }
    
    /* Get index-th item from list... take into account the fact that index needs
       to be increased by one, because first track we'll get with following function
       will be the lead-in... */
    ret_track = g_list_nth_data(_priv->tracks_list, index + 1);
    
    if (ret_track) {
        /* Return track to user if she wants it */
        if (track) {
            g_object_ref(ret_track);
            *track = ret_track;
        }
        return TRUE;
    }
    
    mirage_error(MIRAGE_E_TRACKNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_session_get_track_by_number (MIRAGE_Session *self, gint track_number, GObject **track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *ret_track = NULL;
    GList *entry = NULL;
    
    /* Go over all tracks */
    G_LIST_FOR_EACH(entry, _priv->tracks_list) {        
        gint cur_number = 0;
        
        ret_track = entry->data;
        
        mirage_track_layout_get_track_number(MIRAGE_TRACK(ret_track), &cur_number, NULL);
        
        /* Break the loop if number matches */
        if (track_number == cur_number) {
            break;
        } else {
            ret_track = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_track) {
        mirage_error(MIRAGE_E_TRACKNOTFOUND, error);
        return FALSE;
    }
    
    /* Return track to user if she wants it */
    if (track) {
        g_object_ref(ret_track);
        *track = ret_track;
    }
    
    return TRUE;
}

static gboolean __mirage_session_get_track_by_address (MIRAGE_Session *self, gint address, GObject **track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *ret_track = NULL;
    GList *entry = NULL;
    
    if ((address < _priv->start_sector) || (address >= (_priv->start_sector + _priv->length))) {
        mirage_error(MIRAGE_E_SECTOROUTOFRANGE, error);
        return FALSE;
    }
    
    /* Go over all tracks */
    G_LIST_FOR_EACH(entry, _priv->tracks_list) {        
        gint start_sector = 0;
        gint length = 0;
        
        ret_track = entry->data;
        
        mirage_track_layout_get_start_sector(MIRAGE_TRACK(ret_track), &start_sector, NULL);
        mirage_track_layout_get_length(MIRAGE_TRACK(ret_track), &length, NULL);
        
        /* Break the loop if address lies within track boundaries */
        if (address >= start_sector && address < start_sector + length) {
            break;
        } else {
            ret_track = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_track) {
        mirage_error(MIRAGE_E_TRACKNOTFOUND, error);
        return FALSE;
    }
    
    /* Return track to user if she wants it */
    if (track) {
        g_object_ref(ret_track);
        *track = ret_track;
    }
    
    return TRUE;
}

static gboolean __mirage_session_for_each_track (MIRAGE_Session *self, GFunc func, gpointer user_data, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GList *entry = NULL;
    
    CHECK_ARG(func);
    
    G_LIST_FOR_EACH(entry, _priv->tracks_list) {
        (*func) (MIRAGE_TRACK(entry->data), user_data);
    }
    
    return TRUE;
}

static gboolean __mirage_session_get_track_before (MIRAGE_Session *self, GObject *cur_track, GObject **prev_track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    gint index = 0;
    
    CHECK_ARG(cur_track);
    
    /* Get index of given track in the list */
    index = g_list_index(_priv->tracks_list, cur_track);
    if (index == -1) {
        mirage_error(MIRAGE_E_NOTINLAYOUT, error);
        return FALSE;
    }
    index -= 1; /* Because lead-in has index 0... */
    
    /* Now check if we didn't pass the first track (index = 0) and return previous one */
    if (index > 0) {
        return mirage_session_get_track_by_index(self, index - 1, prev_track, error);
    }
    
    mirage_error(MIRAGE_E_TRACKNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_session_get_track_after (MIRAGE_Session *self, GObject *cur_track, GObject **next_track, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    gint num_tracks = 0;
    gint index = 0;
    
    CHECK_ARG(cur_track);

    /* Get index of given track in the list */
    index = g_list_index(_priv->tracks_list, cur_track);
    if (index == -1) {
        mirage_error(MIRAGE_E_NOTINLAYOUT, error);
        return FALSE;
    }
    index -= 1; /* Because lead-in has index 0... */
    
    /* Now check if we didn't pass the last track (index = num_tracks - 1) and return previous one */
    mirage_session_get_number_of_tracks(self, &num_tracks, NULL);
    if (index < num_tracks - 1) {
        return mirage_session_get_track_by_index(self, index + 1, next_track, error);
    }
    
    mirage_error(MIRAGE_E_TRACKNOTFOUND, error);
    return FALSE;
}


static gboolean __mirage_session_get_number_of_languages (MIRAGE_Session *self, gint *number_of_languages, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    CHECK_ARG(number_of_languages);
    /* Return number of languages */
    *number_of_languages = g_list_length(_priv->languages_list); /* Length of list */
    return TRUE;
}

static gboolean __mirage_session_add_language (MIRAGE_Session *self, gint langcode, GObject **language, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *new_language = NULL;
    
    /* Check if language already exists */
    if (mirage_session_get_language_by_code(self, langcode, NULL, NULL)) {
        mirage_error(MIRAGE_E_LANGEXISTS, error);
        return FALSE;
    }
    
    /* If there's language provided, use it; else create new one */
    if (language && *language) {
        new_language = *language;
        /* If language is not MIRAGE_CDText... */
        if (!MIRAGE_IS_LANGUAGE(new_language)) {
            mirage_error(MIRAGE_E_INVALIDARG, error);
            return FALSE;
        }
        g_object_ref(new_language);
    } else {
        new_language = g_object_new(MIRAGE_TYPE_LANGUAGE, NULL);
    }
    
    /* Set language code */
    mirage_language_set_langcode(MIRAGE_LANGUAGE(new_language), langcode, NULL);
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_language), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_language, NULL);
    
    /* Insert language into language list */
    _priv->languages_list = g_list_insert_sorted(_priv->languages_list, new_language, (GCompareFunc)__sort_languages_by_code);
    
    /* Return language to user if she wants it */
    if (language && (*language == NULL)) {
        g_object_ref(new_language);
        *language = new_language;
    }
    
    return TRUE;
}

static gboolean __mirage_session_remove_language_by_index (MIRAGE_Session *self, gint index, GError **error) {
    /*MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);*/
    GObject *language = NULL;
    gboolean succeeded = FALSE;
    
    /* Find track by index */
    if (!mirage_session_get_language_by_index(self, index, &language, error)) {
        return FALSE;
    }
    
    /* Remove track from list */
    succeeded = __remove_language_from_session(self, language, error);
    g_object_unref(language); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_session_remove_language_by_code (MIRAGE_Session *self, gint langcode, GError **error) {
    /*MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);*/
    GObject *language = NULL;
    gboolean succeeded = FALSE;
    
    /* Find session in layout */
    if (!mirage_session_get_language_by_code(self, langcode, &language, error)) {
        return FALSE;
    }
        
    /* Remove track from list */
    succeeded = __remove_language_from_session(self, language, error);
    g_object_unref(language); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_session_remove_language_by_object (MIRAGE_Session *self, GObject *language, GError **error) {
    CHECK_ARG(language);
    return __remove_language_from_session(self, language, error);
}

static gboolean __mirage_session_get_language_by_index (MIRAGE_Session *self, gint index, GObject **language, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *ret_language = NULL;
    gint num_languages = 0;
    
    /* First language, last language... allow negative indexes to go from behind */
    mirage_session_get_number_of_languages(self, &num_languages, NULL);
    if (index < -num_languages || index >= num_languages) {
        mirage_error(MIRAGE_E_INDEXOUTOFRANGE, error);
        return FALSE;
    } else if (index < 0) {
        index += num_languages; 
    }
    
    /* Get index-th item from list... */
    ret_language = g_list_nth_data(_priv->languages_list, index);
    
    if (ret_language) {
        /* Return language to user if she wants it */
        if (language) {
            g_object_ref(ret_language);
            *language = ret_language;
        }
        return TRUE;
    }
    
    mirage_error(MIRAGE_E_LANGNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_session_get_language_by_code (MIRAGE_Session *self, gint langcode, GObject **language, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GObject *ret_language = NULL;
    GList *entry = NULL;
    
    /* Go over all languages */
    G_LIST_FOR_EACH(entry, _priv->languages_list) {        
        gint cur_code = 0;
        
        ret_language = entry->data;
        
        mirage_language_get_langcode(MIRAGE_LANGUAGE(ret_language), &cur_code, NULL);
        
        /* Break the loop if code matches */
        if (langcode == cur_code) {
            break;
        } else {
            ret_language = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_language) {
        mirage_error(MIRAGE_E_LANGNOTFOUND, error);
        return FALSE;
    }
    
    /* Return language to user if she wants it */
    if (language) {
        g_object_ref(ret_language);
        *language = ret_language;
    }
    
    return TRUE;
}

static gboolean __mirage_session_for_each_language (MIRAGE_Session *self, GFunc func, gpointer user_data, GError **error) {
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GList *entry = NULL;
    
    CHECK_ARG(func);
    
    G_LIST_FOR_EACH(entry, _priv->languages_list) {
        (*func) (MIRAGE_LANGUAGE(entry->data), user_data);
    }
    
    return TRUE;
}

static gboolean __set_cdtext_data (gint langcode, gint type, gint track, guint8 *data, gint len, MIRAGE_Session *self) {
    g_debug("%s: langcode: %i, type: %i, track: %i, data: <%s>; len: %i\n", __func__, langcode, type, track, data, len);
    gboolean succeeded = FALSE;
    GObject *language = NULL;
    
    if (track == 0) {
        /* Session */
        if (!mirage_session_get_language_by_code(self, langcode, &language, NULL)) {
            /* If language does not exist, create it */
            if (!mirage_session_add_language(self, langcode, &language, NULL)) {
                return FALSE;
            }
        }
    } else {
        /* Track */
        GObject *cur_track = NULL;
        if (!mirage_session_get_track_by_number(self, track, &cur_track, NULL)) {
            return FALSE;
        }
        
        if (!mirage_track_get_language_by_code(MIRAGE_TRACK(cur_track), langcode, &language, NULL)) {
            /* If language does not exist, create it */
            if (!mirage_track_add_language(MIRAGE_TRACK(cur_track), langcode, &language, NULL)) {
                g_object_unref(cur_track);
                return FALSE;
            }
        }
        
        g_object_unref(cur_track);
    }
    
    succeeded = mirage_language_set_pack_data(MIRAGE_LANGUAGE(language), type, (gchar *)data, len, NULL);
    
    g_object_unref(language);
    
    return succeeded;
}

static gboolean __mirage_session_set_cdtext_data (MIRAGE_Session *self, guint8 *data, gint len, GError **error) {
    GObject *decoder = NULL;
    gboolean succeeded = FALSE;
    
    CHECK_ARG(data);
    
    /* Create decoder object and hope it'll do all the dirty work correctly... */
    decoder = g_object_new(MIRAGE_TYPE_CDTEXT_ENCDEC, NULL);
    
    succeeded = mirage_cdtext_decoder_init(MIRAGE_CDTEXT_ENCDEC(decoder), data, len, error);
    if (succeeded) {
        gint i = 0;
        /* Try all possible blocks... */
        while (mirage_cdtext_decoder_get_block_info(MIRAGE_CDTEXT_ENCDEC(decoder), i, NULL, NULL, NULL, NULL)) {
            mirage_cdtext_decoder_get_data(MIRAGE_CDTEXT_ENCDEC(decoder), i, (MCDTextDataCallback)__set_cdtext_data, self, error);
            i++;
        }
    }
    
    g_object_unref(decoder);
    
    return succeeded;
}

static gboolean __mirage_session_get_cdtext_data (MIRAGE_Session *self, guint8 **data, gint *len, GError **error) {
    gint num_languages = 0;
    gint num_tracks = 0;
    gint buflen = 0;
    guint8 *buffer = NULL;
    GObject *encoder = NULL;
    
    /* Allocate space... technically, there could be 255 packs (each 18-byte long) 
       per every language... and spec says we support 8 of those. So we play safe 
       here; get number of languages and go for max number of packs */
    mirage_session_get_number_of_languages(self, &num_languages, NULL);
    mirage_session_get_number_of_tracks(self, &num_tracks, NULL);
    
    buflen = num_languages*255*18 /* Size of CD-TEXT pack */;
    buffer = g_malloc0(buflen);
       
    /* Set up encoder */
    encoder = g_object_new(MIRAGE_TYPE_CDTEXT_ENCDEC, NULL);
    mirage_cdtext_encoder_init(MIRAGE_CDTEXT_ENCDEC(encoder), buffer, buflen, NULL);
    
    /* Supported pack types */
    gint pack_types[] = {
        MIRAGE_LANGUAGE_PACK_TITLE,
        MIRAGE_LANGUAGE_PACK_PERFORMER,
        MIRAGE_LANGUAGE_PACK_SONGWRITER,
        MIRAGE_LANGUAGE_PACK_COMPOSER,
        MIRAGE_LANGUAGE_PACK_ARRANGER,
        MIRAGE_LANGUAGE_PACK_MESSAGE,
        MIRAGE_LANGUAGE_PACK_DISC_ID,
        MIRAGE_LANGUAGE_PACK_GENRE,
        MIRAGE_LANGUAGE_PACK_TOC,
        MIRAGE_LANGUAGE_PACK_TOC2,
        MIRAGE_LANGUAGE_PACK_RES_8A,
        MIRAGE_LANGUAGE_PACK_RES_8B,
        MIRAGE_LANGUAGE_PACK_RES_8C,
        MIRAGE_LANGUAGE_PACK_CLOSED_INFO,
        MIRAGE_LANGUAGE_PACK_UPC_ISRC,
        /*MIRAGE_LANGUAGE_PACK_SIZE*/
    };
    gint i, j, k;
    
    /* Add all languages' data to encoder */
    for (i = 0; i < num_languages; i++) {
        GObject* session_lang = NULL;
        gint langcode = 0;
        
        if (!mirage_session_get_language_by_index(self, i, &session_lang, error)) {
            g_object_unref(encoder);
            g_free(buffer);
            return FALSE;
        }
        
        mirage_language_get_langcode(MIRAGE_LANGUAGE(session_lang), &langcode, NULL);
        mirage_cdtext_encoder_set_block_info(MIRAGE_CDTEXT_ENCDEC(encoder), i, langcode, 0, 0, NULL);
        
        /* Pack all supported pack types */
        for (j = 0; j < G_N_ELEMENTS(pack_types); j++) {
            gint pack_type = pack_types[j];
            
            guint8 *session_lang_data = NULL;
            gint session_lang_len = 0;
            
            if (mirage_language_get_pack_data(MIRAGE_LANGUAGE(session_lang), pack_type, (gchar **)&session_lang_data, &session_lang_len, NULL)) {
                /*g_debug("%s: Session; pack type: %02Xh; pack len: %i; pack data: <%s>\n", __func__, pack_type, session_lang_len, session_lang_data);*/
                mirage_cdtext_encoder_add_data(MIRAGE_CDTEXT_ENCDEC(encoder), langcode, pack_type, 0, session_lang_data, session_lang_len, NULL);
            }
            
            g_free(session_lang_data);
            
            /* Now get and pack the same data for the all tracks */
            for (k = 0; k < num_tracks; k++) {
                GObject *track = NULL;
                gint number = 0;
                
                GObject *track_lang = NULL;
                guint8 *track_lang_data = NULL;
                gint track_lang_len = 0;
                
                mirage_session_get_track_by_index(self, k, &track, NULL);
                mirage_track_layout_get_track_number(MIRAGE_TRACK(track), &number, NULL);
                
                /*g_debug("%s: Track %i...\n", __func__, number);*/
                
                if (!mirage_track_get_language_by_code(MIRAGE_TRACK(track), langcode, &track_lang, NULL)) {
                    /*g_debug("%s: Failed to get language with code %i on track %i!\n", __func__, langcode, number);*/
                    continue;
                }
                
                if (mirage_language_get_pack_data(MIRAGE_LANGUAGE(track_lang), pack_type, (gchar **)&track_lang_data, &track_lang_len, NULL)) {
                    /*g_debug("%s: Track %i; pack type: %02Xh; pack len: %i; pack data: <%s>\n", __func__, number, pack_type, track_lang_len, track_lang_data);*/
                    mirage_cdtext_encoder_add_data(MIRAGE_CDTEXT_ENCDEC(encoder), langcode, pack_type, number, track_lang_data, track_lang_len, NULL);
                }
                   
                g_object_unref(track_lang);
                g_object_unref(track);
            }
            
        }
        
        g_object_unref(session_lang);
    }

    /* Encode */
    return mirage_cdtext_encoder_encode(MIRAGE_CDTEXT_ENCDEC(encoder), data, len, error);    
}


static gboolean __mirage_session_get_prev (MIRAGE_Session *self, GObject **prev_session, GError **error) {
    GObject *disc = NULL;
    gboolean succeeded = TRUE;
    
    /* Get parent disc */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(self), &disc, error)) {
        return FALSE;
    }
    
    succeeded = mirage_disc_get_session_before(MIRAGE_DISC(disc), G_OBJECT(self), prev_session, error);
    g_object_unref(disc);
    
    return succeeded;
}

static gboolean __mirage_session_get_next (MIRAGE_Session *self, GObject **next_session, GError **error) {
    GObject *disc = NULL;
    gboolean succeeded = TRUE;
   
    /* Get parent disc */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(self), &disc, error)) {
        return FALSE;
    }
    
    succeeded = mirage_disc_get_session_before(MIRAGE_DISC(disc), G_OBJECT(self), next_session, error);
    g_object_unref(disc);
    
    return succeeded;
}


/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
gboolean mirage_session_set_session_type (MIRAGE_Session *self, gint type, GError **error) {
    SANITY_CHECK(set_session_type);
    return MIRAGE_SESSION_GET_CLASS(self)->set_session_type(self, type, error);
}

gboolean mirage_session_get_session_type (MIRAGE_Session *self, gint *type, GError **error) {
    SANITY_CHECK(get_session_type);
    return MIRAGE_SESSION_GET_CLASS(self)->get_session_type(self, type, error);
}


gboolean mirage_session_layout_set_session_number (MIRAGE_Session *self, gint number, GError **error) {
    SANITY_CHECK(layout_set_session_number);
    return MIRAGE_SESSION_GET_CLASS(self)->layout_set_session_number(self,number, error);
}

gboolean mirage_session_layout_get_session_number (MIRAGE_Session *self, gint *number, GError **error) {
    SANITY_CHECK(layout_get_session_number);
    return MIRAGE_SESSION_GET_CLASS(self)->layout_get_session_number(self, number, error);
}

gboolean mirage_session_layout_set_first_track (MIRAGE_Session *self, gint first_track, GError **error) {
    SANITY_CHECK(layout_set_first_track);
    return MIRAGE_SESSION_GET_CLASS(self)->layout_set_first_track(self, first_track, error);
}

gboolean mirage_session_layout_get_first_track (MIRAGE_Session *self, gint *first_track, GError **error) {
    SANITY_CHECK(layout_get_first_track);
    return MIRAGE_SESSION_GET_CLASS(self)->layout_get_first_track(self, first_track, error);
}

gboolean mirage_session_layout_set_start_sector (MIRAGE_Session *self, gint start_sector, GError **error) {
    SANITY_CHECK(layout_set_start_sector);
    return MIRAGE_SESSION_GET_CLASS(self)->layout_set_start_sector(self, start_sector, error);
}

gboolean mirage_session_layout_get_start_sector (MIRAGE_Session *self, gint *start_sector, GError **error) {
    SANITY_CHECK(layout_get_start_sector);
    return MIRAGE_SESSION_GET_CLASS(self)->layout_get_start_sector(self, start_sector, error);
}

gboolean mirage_session_layout_get_length (MIRAGE_Session *self, gint *length, GError **error) {
    SANITY_CHECK(layout_get_length);
    return MIRAGE_SESSION_GET_CLASS(self)->layout_get_length(self, length, error);
}


gboolean mirage_session_get_number_of_tracks (MIRAGE_Session *self, gint *number_of_tracks, GError **error) {
    SANITY_CHECK(get_number_of_tracks);
    return MIRAGE_SESSION_GET_CLASS(self)->get_number_of_tracks(self, number_of_tracks, error);
}

gboolean mirage_session_add_track_by_index (MIRAGE_Session *self, gint index, GObject **track, GError **error) {
    SANITY_CHECK(add_track_by_index);
    return MIRAGE_SESSION_GET_CLASS(self)->add_track_by_index(self, index, track, error);
}

gboolean mirage_session_add_track_by_number (MIRAGE_Session *self, gint number, GObject **track, GError **error) {
    SANITY_CHECK(add_track_by_number);
    return MIRAGE_SESSION_GET_CLASS(self)->add_track_by_number(self, number, track, error);
}

gboolean mirage_session_remove_track_by_index (MIRAGE_Session *self, gint index, GError **error) {
    SANITY_CHECK(remove_track_by_index);
    return MIRAGE_SESSION_GET_CLASS(self)->remove_track_by_index(self, index, error);
}

gboolean mirage_session_remove_track_by_number (MIRAGE_Session *self, gint number, GError **error) {
    SANITY_CHECK(remove_track_by_number);
    return MIRAGE_SESSION_GET_CLASS(self)->remove_track_by_number(self, number, error);
}

gboolean mirage_session_remove_track_by_object (MIRAGE_Session *self, GObject *track, GError **error) {
    SANITY_CHECK(remove_track_by_object);
    return MIRAGE_SESSION_GET_CLASS(self)->remove_track_by_object(self, track, error);
}

gboolean mirage_session_get_track_by_index (MIRAGE_Session *self, gint index, GObject **track, GError **error) {
    SANITY_CHECK(get_track_by_index);
    return MIRAGE_SESSION_GET_CLASS(self)->get_track_by_index(self, index, track, error);
}

gboolean mirage_session_get_track_by_number (MIRAGE_Session *self, gint number, GObject **track, GError **error) {
    SANITY_CHECK(get_track_by_number);
    return MIRAGE_SESSION_GET_CLASS(self)->get_track_by_number(self, number, track, error);
}

gboolean mirage_session_get_track_by_address (MIRAGE_Session *self, gint address, GObject **track, GError **error) {
    SANITY_CHECK(get_track_by_address);
    return MIRAGE_SESSION_GET_CLASS(self)->get_track_by_address(self, address, track, error);
}

gboolean mirage_session_for_each_track (MIRAGE_Session *self, GFunc func, gpointer user_data, GError **error) {
    SANITY_CHECK(for_each_track);
    return MIRAGE_SESSION_GET_CLASS(self)->for_each_track(self, func, user_data, error);
}

gboolean mirage_session_get_track_before (MIRAGE_Session *self, GObject *cur_track, GObject **prev_track, GError **error) {
    SANITY_CHECK(get_track_before);
    return MIRAGE_SESSION_GET_CLASS(self)->get_track_before(self, cur_track, prev_track, error);
}

gboolean mirage_session_get_track_after (MIRAGE_Session *self, GObject *cur_track, GObject **next_track, GError **error) {
    SANITY_CHECK(get_track_after);
    return MIRAGE_SESSION_GET_CLASS(self)->get_track_after(self, cur_track, next_track, error);
}
    

gboolean mirage_session_get_number_of_languages (MIRAGE_Session *self, gint *number_of_languages, GError **error) {
    SANITY_CHECK(get_number_of_languages);
    return MIRAGE_SESSION_GET_CLASS(self)->get_number_of_languages(self, number_of_languages, error);
}

gboolean mirage_session_add_language (MIRAGE_Session *self, gint langcode, GObject **language, GError **error) {
    SANITY_CHECK(add_language);
    return MIRAGE_SESSION_GET_CLASS(self)->add_language(self, langcode, language, error);
}

gboolean mirage_session_remove_language_by_index (MIRAGE_Session *self, gint index, GError **error) {
    SANITY_CHECK(remove_language_by_index);
    return MIRAGE_SESSION_GET_CLASS(self)->remove_language_by_index(self, index, error);
}

gboolean mirage_session_remove_language_by_code (MIRAGE_Session *self, gint langcode, GError **error) {
    SANITY_CHECK(remove_language_by_code);
    return MIRAGE_SESSION_GET_CLASS(self)->remove_language_by_code(self, langcode, error);
}

gboolean mirage_session_remove_language_by_object (MIRAGE_Session *self, GObject *language, GError **error) {
    SANITY_CHECK(remove_language_by_object);
    return MIRAGE_SESSION_GET_CLASS(self)->remove_language_by_object(self, language, error);
}

gboolean mirage_session_get_language_by_index (MIRAGE_Session *self, gint index, GObject **language, GError **error) {
    SANITY_CHECK(get_language_by_index);
    return MIRAGE_SESSION_GET_CLASS(self)->get_language_by_index(self, index, language, error);
}

gboolean mirage_session_get_language_by_code (MIRAGE_Session *self, gint langcode, GObject **language, GError **error) {
    SANITY_CHECK(get_language_by_code);
    return MIRAGE_SESSION_GET_CLASS(self)->get_language_by_code(self, langcode, language, error);
}

gboolean mirage_session_for_each_language (MIRAGE_Session *self, GFunc func, gpointer user_data, GError **error) {
    SANITY_CHECK(for_each_language);
    return MIRAGE_SESSION_GET_CLASS(self)->for_each_language(self, func, user_data, error);
}

gboolean mirage_session_set_cdtext_data (MIRAGE_Session *self, guint8 *data, gint len, GError **error) {
    SANITY_CHECK(set_cdtext_data);
    return MIRAGE_SESSION_GET_CLASS(self)->set_cdtext_data(self, data, len, error);
}

gboolean mirage_session_get_cdtext_data (MIRAGE_Session *self, guint8 **data, gint *len, GError **error) {
    SANITY_CHECK(get_cdtext_data);
    return MIRAGE_SESSION_GET_CLASS(self)->get_cdtext_data(self, data, len, error);
}


/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static MIRAGE_ObjectClass *parent_class = NULL;

static void __mirage_session_instance_init (GTypeInstance *instance, gpointer g_class) {
    MIRAGE_Session *self = MIRAGE_SESSION(instance);
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    
    _priv->session_number = 1;
    _priv->first_track = 1;
    
    /* Create lead-in */
    mirage_session_add_track_by_number(self, MIRAGE_TRACK_LEADIN, NULL, NULL);
    
    /* Create lead-out */
    mirage_session_add_track_by_number(self, MIRAGE_TRACK_LEADOUT, NULL, NULL);
    
    return;
}

static void __mirage_session_dispose (GObject *obj) {
    MIRAGE_Session *self = MIRAGE_SESSION(obj);
    MIRAGE_SessionPrivate *_priv = MIRAGE_SESSION_GET_PRIVATE(self);
    GList *entry = NULL;
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }

    /* Free list of tracks */
    G_LIST_FOR_EACH(entry, _priv->tracks_list) {
        if (entry->data) {
            GObject *track = entry->data;
            /* Disconnect signal handler and unref */
            g_signal_handlers_disconnect_by_func(MIRAGE_OBJECT(track), __track_modified_handler, self);
            g_object_unref(track);
        }
    }
    g_list_free(_priv->tracks_list);
    
    /* Free list of languages */
    G_LIST_FOR_EACH(entry, _priv->languages_list) {
        if (entry->data) {
            GObject *language = entry->data;
            g_object_unref(language);
        }
    }
    g_list_free(_priv->languages_list);
        
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_session_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_SessionClass *klass = MIRAGE_SESSION_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(MIRAGE_SessionPrivate));
    
    /* Initialize GObject members */
    gobject_class->dispose = __mirage_session_dispose;
    
    /* Initialize MIRAGE_Session members */   
    klass->set_session_type = __mirage_session_set_session_type;
    klass->get_session_type = __mirage_session_get_session_type;
        
    klass->layout_set_session_number = __mirage_session_layout_set_session_number;
    klass->layout_get_session_number = __mirage_session_layout_get_session_number;
    klass->layout_set_first_track = __mirage_session_layout_set_first_track;
    klass->layout_get_first_track = __mirage_session_layout_get_first_track;
    klass->layout_set_start_sector = __mirage_session_layout_set_start_sector;
    klass->layout_get_start_sector = __mirage_session_layout_get_start_sector;
    klass->layout_get_length = __mirage_session_layout_get_length;

    klass->get_number_of_tracks = __mirage_session_get_number_of_tracks;
    klass->add_track_by_index = __mirage_session_add_track_by_index;
    klass->add_track_by_number = __mirage_session_add_track_by_number;
    klass->remove_track_by_index = __mirage_session_remove_track_by_index;
    klass->remove_track_by_number = __mirage_session_remove_track_by_number;
    klass->remove_track_by_object = __mirage_session_remove_track_by_object;
    klass->get_track_by_address = __mirage_session_get_track_by_address;
    klass->get_track_by_index = __mirage_session_get_track_by_index;
    klass->get_track_by_number = __mirage_session_get_track_by_number;
    klass->for_each_track = __mirage_session_for_each_track;
    klass->get_track_before = __mirage_session_get_track_before;
    klass->get_track_after = __mirage_session_get_track_after;
    
    klass->get_number_of_languages = __mirage_session_get_number_of_languages;
    klass->add_language = __mirage_session_add_language;
    klass->remove_language_by_index = __mirage_session_remove_language_by_index;
    klass->remove_language_by_code = __mirage_session_remove_language_by_code;
    klass->remove_language_by_object = __mirage_session_remove_language_by_object;
    klass->get_language_by_index = __mirage_session_get_language_by_index;
    klass->get_language_by_code = __mirage_session_get_language_by_code;
    klass->for_each_language = __mirage_session_for_each_language;
    
    klass->set_cdtext_data = __mirage_session_set_cdtext_data;
    klass->get_cdtext_data = __mirage_session_get_cdtext_data;
    
    klass->get_prev = __mirage_session_get_prev;
    klass->get_next = __mirage_session_get_next;
    
    return;
}

GType mirage_session_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_SessionClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_session_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Session),
            0,      /* n_preallocs */
            __mirage_session_instance_init    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_OBJECT, "MIRAGE_Session", &info, 0);
    }
    
    return type;
}
