/*
 *  libMirage: ISO image parser: Disc object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "image-iso.h"

/******************************************************************************\
 *                        Load and save implementation                        *
\******************************************************************************/
static gboolean __mirage_disc_iso_get_image_type (MIRAGE_Disc *self, gchar **image_type, GError **error) {
    *image_type = g_strdup(mirage_plugin_info.image_id);
    return TRUE;
}

static gboolean __mirage_disc_iso_load_track (MIRAGE_Disc *self, gchar *filename, GError **error) {
    gboolean succeeded = TRUE;
    GObject *session = NULL;
    GObject *track = NULL;
    
    if (!g_file_test(filename, G_FILE_TEST_IS_REGULAR)) {
        mirage_error(MIRAGE_E_DATAFILE, error);
        return FALSE;
    }
    
    mirage_disc_get_session_by_index(self, -1, &session, NULL);
    
    /* Add track */
    succeeded = mirage_session_add_track_by_index(MIRAGE_SESSION(session), -1, &track, error);
    g_object_unref(session);
    if (!succeeded) {
        return succeeded;
    }
    
    /* Create fragment */
    GObject *mirage = NULL;
    if (!mirage_object_get_mirage(MIRAGE_OBJECT(self), &mirage, error)) {
        g_object_unref(track);
        return FALSE;
    }
    
    GObject *data_fragment = NULL;
    
    /* Determine whether we have Mode 1 (ISO) or Audio (WAV) file */
    if (mirage_helper_has_suffix(filename, ".iso")
        || mirage_helper_has_suffix(filename, ".img")) {
        /* Mode 1/Mode 2 Form 1... BINARY fragment interface is used */
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: data file, using BINARY fragment interface\n", __func__);

        mirage_mirage_create_fragment(MIRAGE_MIRAGE(mirage), FRAGMENT_IFACE_BINARY, filename, &data_fragment, error);
        g_object_unref(mirage);
        if (!data_fragment) {
            g_object_unref(track);
            return FALSE;
        }
        
        /* Set track file */
        mirage_finterface_binary_track_file_set_handle(MIRAGE_FINTERFACE_BINARY(data_fragment), g_fopen(filename, "r"), NULL);
        mirage_finterface_binary_track_file_set_sectsize(MIRAGE_FINTERFACE_BINARY(data_fragment), 2048, NULL);
        mirage_finterface_binary_track_file_set_format(MIRAGE_FINTERFACE_BINARY(data_fragment), FR_BIN_TFILE_DATA, NULL);        
        
        /* Set track mode */
        mirage_track_set_mode(MIRAGE_TRACK(track), MIRAGE_MODE_MODE1, NULL);
    } else {
        /* Assume audio... AUDIO fragment interface is used */
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: audio file, using AUDIO fragment interface\n", __func__);
        
        mirage_mirage_create_fragment(MIRAGE_MIRAGE(mirage), FRAGMENT_IFACE_AUDIO, filename, &data_fragment, error);
        g_object_unref(mirage);
        if (!data_fragment) {
            g_object_unref(track);
            return FALSE;
        }
        
        /* Set file */
        if (!mirage_finterface_audio_set_file(MIRAGE_FINTERFACE_AUDIO(data_fragment), filename, error)) {
            MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: failed to set file!\n", __func__);
            g_object_unref(data_fragment);
            g_object_unref(track);
            return FALSE;
        }
        
        /* Set track mode */
        mirage_track_set_mode(MIRAGE_TRACK(track), MIRAGE_MODE_AUDIO, NULL);   
    }
    
    /* Use whole file */
    if (!mirage_fragment_use_the_rest_of_file(MIRAGE_FRAGMENT(data_fragment), error)) {
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: failed to use the rest of file!\n", __func__);
        g_object_unref(data_fragment);
        g_object_unref(track);
        return FALSE;
    }
        
    /* Add fragment */
    if (!mirage_track_add_fragment(MIRAGE_TRACK(track), -1, &data_fragment, error)) {
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: failed to add fragment!\n", __func__);
        g_object_unref(data_fragment);
        g_object_unref(track);
        return FALSE;
    }
        
    g_object_unref(data_fragment);
    
    g_object_unref(track);
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: finished loading track\n", __func__);
    
    return TRUE;    
}

static gboolean __mirage_disc_iso_load_image (MIRAGE_Disc *self, gchar **filenames, GError **error) {
    GObject *session = NULL;
    gint length = 0;
    gint i;
    
    /* Set filenames */
    mirage_disc_set_filenames(self, filenames, NULL);
    
    /* Session: one session (with possibly multiple tracks) */
    if (!mirage_disc_add_session_by_number(self, 1, &session, error)) {
        return FALSE;
    }
    
    /* Data track is always assumed to be Mode 1, and it can be combined with
       audio tracks, so we're more or less getting regular CD-ROM sessions, right? */
    mirage_session_set_session_type(MIRAGE_SESSION(session), MIRAGE_SESSION_CD_ROM, NULL);
    g_object_unref(session);

    /* Load track(s) */
    for (i = 0; i < g_strv_length(filenames); i++) {
        /* Load track(s) */
        if (!__mirage_disc_iso_load_track(self, filenames[i], error)) {
            MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: failed to load track %i!\n", __func__, i);
            return FALSE;
        }
    }
    
    /* Now get length and if it surpasses length of 90min CD, assume we
       have a DVD */
    mirage_disc_layout_get_length(self, &length, NULL);
    if (length > 90*60*75) {
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: DVD-ROM image\n", __func__);
        mirage_disc_set_medium_type(self, MIRAGE_MEDIUM_DVD, NULL);
    } else {
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: CD-ROM image\n", __func__);
        mirage_disc_set_medium_type(self, MIRAGE_MEDIUM_CD, NULL);
        /* If we got CD-ROM, we assume Red Book for now... which means disc starts
           at -150 and first track has 150 sector pregap */
        mirage_disc_layout_set_start_sector(self, -150, NULL);
        
        GObject *ftrack = NULL;
        mirage_disc_get_track_by_index(self, 0, &ftrack, NULL);
        
        /* Add pregap fragment (empty) */
        GObject *mirage = NULL;
        if (!mirage_object_get_mirage(MIRAGE_OBJECT(self), &mirage, error)) {
            g_object_unref(ftrack);
            return FALSE;
        }
        GObject *pregap_fragment = NULL;
        mirage_mirage_create_fragment(MIRAGE_MIRAGE(mirage), FRAGMENT_IFACE_NULL, "NULL", &pregap_fragment, error);
        g_object_unref(mirage);
        if (!pregap_fragment) {
            MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: failed to create pregap fragment\n", __func__);
            g_object_unref(ftrack);
            return FALSE;
        }
        mirage_track_add_fragment(MIRAGE_TRACK(ftrack), 0, &pregap_fragment, NULL);
        mirage_fragment_set_length(MIRAGE_FRAGMENT(pregap_fragment), 150, NULL);
        g_object_unref(pregap_fragment);
        
        /* Track starts at 150 */
        mirage_track_set_track_start(MIRAGE_TRACK(ftrack), 150, NULL);
        
        g_object_unref(ftrack);
    }
    
    return TRUE;
}

static gboolean __mirage_disc_iso_save_image (MIRAGE_Disc *self, MIRAGE_Disc *input_disc, GError **error) {
    mirage_error(MIRAGE_E_NOTIMPL, error);
    return FALSE;
}

/******************************************************************************\
 *                                Object init                                 *
\******************************************************************************/
static void __mirage_disc_iso_class_init (gpointer g_class, gpointer g_class_data) {
    MIRAGE_DiscClass *klass = MIRAGE_DISC_CLASS(g_class);
    
    /* Initialize MIRAGE_Disc members */
    klass->get_image_type = __mirage_disc_iso_get_image_type;
    klass->load_image = __mirage_disc_iso_load_image;
    klass->save_image = __mirage_disc_iso_save_image;
    
    return;
}

GType mirage_disc_iso_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_Disc_ISOClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_disc_iso_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Disc_ISO),
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_DISC, "MIRAGE_Disc_ISO", &info, 0);
    }
    
    return type;
}
