/*
 *  libMirage: MDS image parser: Disc object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "image-mds.h"


/******************************************************************************\
 *                         Disc private functions                             *
\******************************************************************************/
/*
    I hexedited the track mode field with various values and fed it to Alchohol;
    it seemed that high part of byte had no effect at all; only the lower one 
    affected the mode, in the following manner:
    00: Mode 2, 01: Audio, 02: Mode 1, 03: Mode 2, 04: Mode 2 Form 1, 05: Mode 2 Form 2, 06: UKNONOWN, 07: Mode 2
    08: Mode 2, 09: Audio, 0A: Mode 1, 0B: Mode 2, 0C: Mode 2 Form 1, 0D: Mode 2 Form 2, 0E: UKNONOWN, 0F: Mode 2
*/
static gint __mirage_disc_mds_convert_track_mode (MIRAGE_Disc *self, gint mode) {
    /* convert between two values */
    static struct {
        gint mds_mode;
        gint mirage_mode;
    } modes[] = {
        {0x00, MIRAGE_MODE_MODE2},
        {0x01, MIRAGE_MODE_AUDIO},
        {0x02, MIRAGE_MODE_MODE1},
        {0x03, MIRAGE_MODE_MODE2},
        {0x04, MIRAGE_MODE_MODE2_FORM1},
        {0x05, MIRAGE_MODE_MODE2_FORM2},
        /*{0x06, MIRAGE_MODE_UNKNOWN},*/
        {0x07, MIRAGE_MODE_MODE2},
    };
    gint i;
    
    /* basically, do the test twice; once for value, and once for value + 8 */
    for (i = 0; i < G_N_ELEMENTS(modes); i++) {
        if (((mode & 0x0F) == modes[i].mds_mode)
            || ((mode & 0x0F) == modes[i].mds_mode + 8)) {
            return modes[i].mirage_mode;
        }
    }
    
    return -1;
}

static void dump_bytes (guint8 *buf, gint len) {
    gint i;
    for (i = 0; i < len; i++) {
        g_print("%02hhX", buf[i]);
        if (!((i+1) % 8)) 
            g_print("\n");
        else
            g_print(" ");
            
    }
    g_print("\n");
}


/******************************************************************************\
 *                        Load and save implementation                        *
\******************************************************************************/
static gboolean __mirage_disc_mds_get_image_type (MIRAGE_Disc *self, gchar **image_type, GError **error) {
    *image_type = g_strdup(mirage_plugin_info.image_id);
    return TRUE;
}

static gchar *__helper_find_binary_file (gchar *declared_filename, gchar *mds_filename) {
    gchar *bin_fullpath = NULL;
        
    gchar ext[4] = "";
    if (sscanf(declared_filename, "*.%s", ext) == 1) {
        /* Use MDS filename and replace its extension with the one of the data file */
        bin_fullpath = g_strdup(mds_filename);
        gint len = strlen(bin_fullpath);
        sprintf(bin_fullpath+len-3, ext);
    } else {
        gchar *image_path = g_path_get_dirname(mds_filename);
        bin_fullpath = g_strjoin("/", image_path, declared_filename, NULL);
        g_free(image_path);
    }
    
    /* Test */
    if (!g_file_test(bin_fullpath, G_FILE_TEST_IS_REGULAR)) {
        g_free(bin_fullpath);
        bin_fullpath = NULL;
    }
    
    return bin_fullpath;
}

static gboolean __mirage_disc_mds_load_disc (MIRAGE_Disc *self, gchar *filename, MDS_Header *header, FILE *file, GError **error) {
    gint medium_type = 0;
    gint i;
    
    /* We'll need that for filename */
    gint file_length = 0;
    fseeko(file, 0, SEEK_END);
    file_length = ftell(file);
    
    /* We'll need to know medium type */
    mirage_disc_get_medium_type(self, &medium_type, NULL);
    
    /* *** Disc structures *** */
    /* Disc structures: in lead-in areas of DVD and BD discs there are several
       control structures that store various information about the media. There
       are various formats defined in MMC-3 for these structures, and they are 
       retrieved from disc using READ DISC STRUCTURE command. Of all the structures,
       MDS format seems to store only three types: 
        - 0x0001: DVD copyright information (4 bytes)
        - 0x0004: DVD manufacturing information (2048 bytes)
        - 0x0000: Physical format information (2048 bytes)
       They are stored in that order, taking up 4100 bytes. If disc is dual-layer,
       data consists of 8200 bytes, containing afore-mentioned sequence for each
       layer.
       
       -- Rok */
    if (header->disc_structures_offset) {
        MIRAGE_DiscStruct_Copyright copy_info;
        MIRAGE_DiscStruct_Manufacture manu_info;
        MIRAGE_DiscStruct_PhysInfo phys_info;
        
        memset(&copy_info, 0, sizeof(copy_info));
        memset(&manu_info, 0, sizeof(manu_info));
        memset(&phys_info, 0, sizeof(phys_info));
        
        fseeko(file, header->disc_structures_offset, SEEK_SET);
                
        /* DVD copyright information */
        fread(&copy_info, sizeof(copy_info), 1, file);
        
        /* DVD manufacture information */
        fread(&manu_info, sizeof(manu_info), 1, file);
        /*g_debug("     -> DVD Manufacture information (%i)... 2048 bytes of it...\n", sizeof(MIRAGE_DiscStruct_Manufacture));*/
        
        /* Physical information */
        fread(&phys_info, sizeof(phys_info), 1, file);
                
        mirage_disc_set_disc_structure(self, 0, 0x0000, (guint8 *)&phys_info, sizeof(phys_info), NULL);
        mirage_disc_set_disc_structure(self, 0, 0x0001, (guint8 *)&copy_info, sizeof(copy_info), NULL);
        mirage_disc_set_disc_structure(self, 0, 0x0004, (guint8 *)&manu_info, sizeof(manu_info), NULL);
                    
        /* Second round if it's dual-layer... */
        if (phys_info.num_layers == 0x01) {
            memset(&copy_info, 0, sizeof(copy_info));
            memset(&manu_info, 0, sizeof(manu_info));
            memset(&phys_info, 0, sizeof(phys_info));
            
            fread(&copy_info, sizeof(copy_info), 1, file);
            fread(&manu_info, sizeof(manu_info), 1, file);
            fread(&phys_info, sizeof(phys_info), 1, file);
            
            mirage_disc_set_disc_structure(self, 1, 0x0000, (guint8 *)&phys_info, sizeof(phys_info), NULL);
            mirage_disc_set_disc_structure(self, 1, 0x0001, (guint8 *)&copy_info, sizeof(copy_info), NULL);
            mirage_disc_set_disc_structure(self, 1, 0x0004, (guint8 *)&manu_info, sizeof(manu_info), NULL);
        }
    }
    
    /* It seems BCA (Burst Cutting Area) structure is stored as well, but in separate
       place (kinda makes sense, because it doesn't have fixed length) */
    if (header->bca_len) {
        guint8 *bca_data = g_malloc0(header->bca_len);
        
        fseeko(file, header->bca_data_offset, SEEK_SET);
        fread(bca_data, header->bca_len, 1, file);
        
        mirage_disc_set_disc_structure(self, 0, 0x0003, bca_data, header->bca_len, NULL);
        
        g_free(bca_data);
    }
    
    /* *** Sessions *** */
    for (i = 0; i < header->num_sessions; i++) {
        MDS_SessionBlock session;
        memset(&session, 0, sizeof(session));
        
        fseeko(file, header->sessions_blocks_offset + i * sizeof(session), SEEK_SET);
        fread(&session, sizeof(session), 1, file);
        
        /*g_debug("    SESSION BLOCK[%i]:\n", i);        
        g_debug("     - StartAddress: 0x%X\n", session.start_address);
        g_debug("     - Length: 0x%X\n", session.length);
        g_debug("     - SessionNumber: %i\n", session.session_number);
        g_debug("     - NumAllBlocks: %i\n", session.num_all_blocks);
        g_debug("     - NumNonTrackBlocks: %i\n", session.num_nontrack_blocks);
        g_debug("     - FirstTrack: %i\n", session.first_track);
        g_debug("     - LastTrack: %i\n", session.last_track);
        g_debug("     - Dummy2: 0x%X\n", session.__dummy2__);
        g_debug("     - TrackBlocksOffset: 0x%X\n", session.tracks_blocks_offset);*/
        
        GObject *cur_session = NULL;
        if (!mirage_disc_add_session_by_number(self, session.session_number, &cur_session, error)) {
            return FALSE;
        }
        
        /* *** Tracks *** */
        /*g_debug("Loading tracks\n");*/
        gint j;
        for (j = 0; j < session.num_all_blocks; j++) {
            MDS_TrackBlock block;
            memset(&block, 0, sizeof(block));
            
            fseeko(file, session.tracks_blocks_offset + j*sizeof(block), SEEK_SET);
            fread(&block, sizeof(block), 1, file);
            
            /*g_debug("        * TRACK BLOCK[%i]:\n", j);
            g_debug("          - Mode = 0x%X\n", block.mode);
            g_debug("          - Subchannel = 0x%X\n", block.subchannel);
            g_debug("          - AdrCtl = 0x%X\n", block.adr_ctl);
            g_debug("          - Dummy2 = 0x%X\n", block.__dummy2__);
            g_debug("          - Point = 0x%X\n", block.point);
            g_debug("\n");
            g_debug("          - Dummy3: 0x%X\n", block.__dummy3__);
            g_debug("          - Min: 0x%X\n", block.min);
            g_debug("          - Sec: 0x%X\n", block.sec);
            g_debug("          - Frame: 0x%X\n", block.frame);
            g_debug("          - ExtraOffset: 0x%X\n", block.extra_offset);
            g_debug("          - SectorSize: 0x%X\n", block.sector_size);
            g_debug("\n");
            g_debug("          - DummyX4:\n"); //dump_bytes(block.__dummy4__, 18);
            g_debug("          - StartSector: 0x%X\n", block.start_sector);
            g_debug("          - StartOffset: 0x%llX\n", block.start_offset);
            g_debug("          - Session: 0x%X\n", block.session);
            g_debug("          - DummyX5:\n"); //dump_bytes(block.__dummy5__, 3);
            g_debug("          - FooterOffset: 0x%X\n", block.footer_offset);
            g_debug("          - DummyX6:\n"); //dump_bytes(block.__dummy6__, 24);*/
                    
            MDS_TrackExtraBlock extra_block;
            if (block.extra_offset) {
                memset(&extra_block, 0, sizeof(extra_block));
                guint32 position2 = ftell(file);
                fseeko(file, block.extra_offset, SEEK_SET);
                fread(&extra_block, sizeof(extra_block), 1, file);
                fseeko(file, position2, SEEK_SET);
            
                /*g_debug("        * EXTRA TRACK BLOCK[%i]:\n", j);
                g_debug("          - PreGap = 0x%X\n", extra_block.pregap);
                g_debug("          - Length = 0x%X\n", extra_block.length);*/
            }
            
            MDS_Footer footer_block;
            if (block.footer_offset) {
                memset(&footer_block, 0, sizeof(footer_block));
                guint32 position2 = ftell(file);
                fseeko(file, block.footer_offset, SEEK_SET);
                fread(&footer_block, sizeof(footer_block), 1, file);
                fseeko(file, position2, SEEK_SET);
            
                /*g_debug("        * FOOTER BLOCK[%i]:\n", j);
                g_debug("          - Filename offset = 0x%X\n", footer_block.filename_offset);*/
            }
            
            if (block.point > 0 && block.point < 99) {
                /* Track */
                GObject *cur_track = NULL;
                if (!mirage_session_add_track_by_number(MIRAGE_SESSION(cur_session), block.point, &cur_track, error)) {
                    g_object_unref(cur_session);
                    return FALSE;
                }
                    
                gint converted_mode = __mirage_disc_mds_convert_track_mode(self, block.mode);
                mirage_track_set_mode(MIRAGE_TRACK(cur_track), converted_mode, NULL);
                    
                /* Flags: decoded from Ctl */
                gint flags = 0;
                if (block.adr_ctl & 0x8) flags |= MIRAGE_TRACKF_CHAN;
                if (block.adr_ctl & 0x2) flags |= MIRAGE_TRACKF_COPY;
                if (block.adr_ctl & 0x1) flags |= MIRAGE_TRACKF_EMPH;
                mirage_track_set_flags(MIRAGE_TRACK(cur_track), flags, NULL);
                
                /* Track file: it seems all tracks have the same extra block, and
                   that filename is located at the end of it... meaning filename's 
                   length is from filename_offset to end of the file */
                gint filename_length = file_length - footer_block.filename_offset;
                gchar *tmp_mdf_filename = g_malloc0(filename_length);
                fseeko(file, footer_block.filename_offset, SEEK_SET);
                fread(tmp_mdf_filename, filename_length, 1, file);
                gchar *mdf_filename = __helper_find_binary_file(tmp_mdf_filename, filename);
                g_free(tmp_mdf_filename);
                
                if (!mdf_filename) {
                    mirage_error(MIRAGE_E_DATAFILE, error);
                    return FALSE;
                }
                
                /* Get Mirage and have it make us binary fragment */
                GObject *mirage = NULL;
    
                if (!mirage_object_get_mirage(MIRAGE_OBJECT(self), &mirage, error)) {
                    g_object_unref(cur_track);
                    g_object_unref(cur_session);
                    return FALSE;
                }
    
                GObject *data_fragment = NULL;
                mirage_mirage_create_fragment(MIRAGE_MIRAGE(mirage), FRAGMENT_IFACE_BINARY, mdf_filename, &data_fragment, error);
                g_object_unref(mirage);
                if (!data_fragment) {
                    g_object_unref(cur_track);
                    g_object_unref(cur_session);
                    return FALSE;
                }
            
                /* Prepare data fragment */
                FILE *tfile_handle = g_fopen(mdf_filename, "r");
                guint64 tfile_offset = block.start_offset;
                gint tfile_sectsize = block.sector_size;
                gint tfile_format = 0;
                
                if (converted_mode == MIRAGE_MODE_AUDIO) {
                    tfile_format = FR_BIN_TFILE_AUDIO;
                } else {
                    tfile_format = FR_BIN_TFILE_DATA;
                }
                
                gint fragment_len = 0;
                
                g_free(mdf_filename);
                
                if (medium_type == MIRAGE_MEDIUM_DVD) {
                    /* Length: length of DVD-ROM's track is same as the length of the session */
                    fragment_len = session.length;
                } else {
                    /* Length: for CD-ROMs, track lengths are stored in extra blocks */
                    fragment_len = extra_block.length;
                }
                
                mirage_fragment_set_length(MIRAGE_FRAGMENT(data_fragment), fragment_len, NULL);
                
                mirage_finterface_binary_track_file_set_handle(MIRAGE_FINTERFACE_BINARY(data_fragment), tfile_handle, NULL);
                mirage_finterface_binary_track_file_set_offset(MIRAGE_FINTERFACE_BINARY(data_fragment), tfile_offset, NULL);
                mirage_finterface_binary_track_file_set_sectsize(MIRAGE_FINTERFACE_BINARY(data_fragment), tfile_sectsize, NULL);
                mirage_finterface_binary_track_file_set_format(MIRAGE_FINTERFACE_BINARY(data_fragment), tfile_format, NULL);
                
                /* Subchannel */
                switch (block.subchannel) {
                    case MDS_SUBCHAN_PW_INTERLEAVED: {
                        gint sfile_sectsize = 96;
                        gint sfile_format = FR_BIN_SFILE_PW96_INT | FR_BIN_SFILE_INT;
                                                
                        mirage_finterface_binary_subchannel_file_set_sectsize(MIRAGE_FINTERFACE_BINARY(data_fragment), sfile_sectsize, NULL);
                        mirage_finterface_binary_subchannel_file_set_format(MIRAGE_FINTERFACE_BINARY(data_fragment), sfile_format, NULL);
                        
                        /* We need to correct the data for track sector size...
                           MDS format has already added 96 bytes to sector size,
                           so we need to subtract it */
                        tfile_sectsize = block.sector_size - sfile_sectsize;
                        mirage_finterface_binary_track_file_set_sectsize(MIRAGE_FINTERFACE_BINARY(data_fragment), tfile_sectsize, NULL);
                            
                        break;
                    }
                }
                
                mirage_track_add_fragment(MIRAGE_TRACK(cur_track), -1, &data_fragment, error);
                g_object_unref(data_fragment);
                
                g_object_unref(cur_track);
            }
        }
        
        g_object_unref(cur_session);
    }

    return TRUE;
}


static gboolean __mirage_disc_mds_load_image (MIRAGE_Disc *self, gchar **filenames, GError **error) {
    gboolean succeeded = TRUE;
    
    /* For now, MDS parser supports only one-file images */
    if (g_strv_length(filenames) > 1) {
        mirage_error(MIRAGE_E_SINGLEFILE, error);
        return FALSE;
    }
    
    /* Open file */
    FILE *file = g_fopen(filenames[0], "r");
    if (!file) {
        mirage_error(MIRAGE_E_IMAGEFILE, error);
        return FALSE;
    }
    
    mirage_disc_set_filenames(self, filenames, NULL);
    
    MDS_Header header;
    memset(&header, 0, sizeof(header));
    fread(&header, sizeof(header), 1, file);
        
    /*g_debug("HEADER:\n");
    g_debug(" - Signature: %.16s\n", header.signature);
    g_debug(" - Version: %u.%u\n", header.version[0], header.version[1]);
    g_debug(" - Media Type: 0x%X\n", header.disc_media);
    g_debug(" - Number of Sessions: 0x%X\n", header.num_sessions);
    g_debug(" - Dummy1: 0x%X, 0x%X\n", header.__dummy1__[0], header.__dummy1__[1]);
    g_debug(" - BCA Length: 0x%X\n", header.bca_len);
    g_debug(" - Dummy2: 0x%X, 0x%X\n", header.__dummy2__[0], header.__dummy2__[1]);
    g_debug(" - BCA Data Offset: 0x%X\n", header.bca_data_offset);
    g_debug(" - Dummy3: 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X\n", header.__dummy3__[0], header.__dummy3__[1], header.__dummy3__[2], header.__dummy3__[3], header.__dummy3__[4], header.__dummy3__[5]);
    g_debug(" - Disc Structures Offset: 0x%X\n", header.disc_structures_offset);
    g_debug(" - Dummy4: 0x%X, 0x%X, 0x%X\n", header.__dummy4__[0], header.__dummy4__[1], header.__dummy4__[2]);
    g_debug(" - Session Blocks Offset: 0x%X\n", header.sessions_blocks_offset);
    g_debug(" - DPM Blocks Offset: 0x%X\n", header.dpm_blocks_offset);*/
    
    switch (header.disc_media) {
        case MDS_DISCMEDIA_CD:
        case MDS_DISCMEDIA_CD_R:
        case MDS_DISCMEDIA_CD_RW: {
            mirage_disc_set_medium_type(self, MIRAGE_MEDIUM_CD, NULL);
            succeeded = __mirage_disc_mds_load_disc(self, filenames[0], &header, file, error);
            break;
        }
        case MDS_DISCMEDIA_DVD: 
        case MDS_DISCMEDIA_DVD_MINUS_R: {
            mirage_disc_set_medium_type(self, MIRAGE_MEDIUM_DVD, NULL);
            succeeded = __mirage_disc_mds_load_disc(self, filenames[0], &header, file, error);            
            break;
        }
        default: {
            MIRAGE_DEBUG(self, MIRAGE_DEBUG_PARSER, "%s: Disc Media of type 0x%X not supported yet!\n", __func__, header.disc_media);
            mirage_error(MIRAGE_E_NOTIMPL, error);
            succeeded = FALSE;
            break;
        }
    }
    
    fclose(file);
        
    return succeeded;    
}

static gboolean __mirage_disc_mds_save_image (MIRAGE_Disc *self, MIRAGE_Disc *input_disc, GError **error) {
    mirage_error(MIRAGE_E_NOTIMPL, error);
    return FALSE;
}


/******************************************************************************\
 *                                Object init                                 *
\******************************************************************************/
static void __mirage_disc_mds_class_init (gpointer g_class, gpointer g_class_data) {
    MIRAGE_DiscClass *klass = MIRAGE_DISC_CLASS(g_class);
    
    /* Initialize MIRAGE_Disc members */
    klass->get_image_type = __mirage_disc_mds_get_image_type;
    klass->load_image = __mirage_disc_mds_load_image;
    klass->save_image = __mirage_disc_mds_save_image;
    
    return;
}

GType mirage_disc_mds_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_Disc_MDSClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_disc_mds_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Disc_MDS),
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_DISC, "MIRAGE_Disc_MDS", &info, 0);
    }
    
    return type;
}
